* = $0300

; Hardware/driver routines for BMOW console video within MSBASIC
;---------------------------------------------

; --- memory-mapped addresses
; VRAM is in bank 20, addresses $0000-$7FFF
VIDEO_RAM_BANK = $20
VIDEO_RAM_BASE = $0000
VIDEO_RAM_A0_L = $7FFC
VIDEO_RAM_A0_H = $7FFD
VIDEO_RAM_A0 = $7FFC
VIDEO_RAM_A1_L = $7FFA
VIDEO_RAM_A1_H = $7FFB
VIDEO_RAM_A1 = $7FFA
VIDEO_RAM_LEFT_MARGIN = $7FF9
VIDEO_RAM_TOP_MARGIN = $7FF8
VIDEO_RAM_CURSOR_PTR_L = $7FFE
VIDEO_RAM_CURSOR_PTR_H = $7FFF
VIDEO_RAM_CURSOR_PTR = $7FFE

CONSOLE_BACKSPACE = 8
CONSOLE_LINE_FEED = 10
CONSOLE_RETURN = 13

CONSOLE_ADR_A0:
CONSOLE_ADR_A0_L: 
!fi 1

CONSOLE_ADR_A0_H: 
!fi 1
 
 
; VSIZE register (write only)

; $00 - 200 lines, mode byte per line
; $40 - 240 lines, mode byte per frame
; $80 - 400 lines, mode byte per line
; $C0 - 480 lines, mode byte per frame
VIDEO_ADR_VSIZE = $003F30

; VBLANK register (read only)
; $80 = drawing frame, $00 = vblank interval
VIDEO_ADR_VBLANK = $003F30

; PALETTE registers (write only)
; palette entry to modify ($00-$FF)
VIDEO_ADR_PAL_INDEX = $003F40

; new RGB value for the entry 
;(must write R,G,B on consecutive writes, only the lowest 6 bits per channel are used)
VIDEO_ADR_PAL_COLOR = $003F44

; mask for pixel data
; use $FF for 8bpp and 4bpp
; use $F3 for 2bpp
; use $F1 for 1bpp
; or setup the palette entries so that the masked bits are "don't care"
VIDEO_ADR_PAL_MASK = $003F48


; --- Console enable text-only mode
; @noparams
console_enable_text_only_mode:
	phb
	bank #VIDEO_RAM_BANK
	clr VIDEO_RAM_LEFT_MARGIN
	clr VIDEO_RAM_CURSOR_PTR_L
	plb
	jsr console_cursor_on
	rts
	
	
; --- Console enable mixed mode
; @noparams
console_enable_mixed_mode:
	phb
	bank #VIDEO_RAM_BANK
	lda #$02
	sta VIDEO_RAM_LEFT_MARGIN
	sta VIDEO_RAM_CURSOR_PTR_L
	plb
	jsr console_cursor_on
	rts
	

; --- Console set row
; @param new row number
console_set_row:
	pha
	jsr console_cursor_off
	pla
	
	phb
	bank #VIDEO_RAM_BANK
	asl
	asl
	sta VIDEO_RAM_CURSOR_PTR_H
	sta VIDEO_RAM_TOP_MARGIN
	plb
	
	jsr console_cursor_on
	rts
	
			
; --- Console standard init
; @noparams
; Clears the screen, and homes the cursor.
console_standard_init:
	; set the video resolution
	lda #$40 ; 240 lines, mode byte per frame = 64x30 text
	sta VIDEO_ADR_VSIZE 
	
	; set the video mode
	lda #$4F ; text, 1bpp, palette 15
	phb
	bank #VIDEO_RAM_BANK
	sta $7800 
	sta $7A80
	sta $7A81
	
	; set the left margin
	clr VIDEO_RAM_LEFT_MARGIN
	
	; set the top margin
	clr VIDEO_RAM_TOP_MARGIN
	
	plb
	
	; set the video mask
	lda #$FF
	sta VIDEO_ADR_PAL_MASK
	
	jsr console_setup_palette
	jsr console_clear_screen
	jsr console_home_cursor
	jsr console_cursor_on
	
	rts
	
		
; --- Console setup palette
; @noparams
; Sets the RGB value of palette entries 240-255 to be
; foreground color if the least significant bit is 1, background if it's 0.

console_setup_palette:
	ldx #$F0
	
-	; background color
	txa
	sta VIDEO_ADR_PAL_INDEX
	
	; black background
	lda #$00
	sta VIDEO_ADR_PAL_COLOR
	sta VIDEO_ADR_PAL_COLOR
	sta VIDEO_ADR_PAL_COLOR
	
	inx
	
	; foreground color
	txa
	sta VIDEO_ADR_PAL_INDEX
	
	; foreground.rgb
	lda #$3F
	sta VIDEO_ADR_PAL_COLOR
	sta VIDEO_ADR_PAL_COLOR
	sta VIDEO_ADR_PAL_COLOR
		
	inx
	
	bne -
					
	rts	
	
	
; --- Console setup multi palette
; @noparams
; Sets the RGB value of all 256 palette entries to be
; foreground color if the least significant bit is 1, background if it's 0.
; palette colors
;    0    white on black
;    1	  white on white
;    2    yellow on black
;    3    yellow on white
;	 4    purple on black
;    5    purple on white
;    6    red on black
;    7    red on white
;    8    cyan on black
;    9    cyan on white
;    A    green on black
;    B    green on white
;    C    blue on black
;    D    blue on white
;    E    black on black
;    F    black on white

console_setup_multi_palette:
	ldx #$00
	
-	; background color
	txa
	sta VIDEO_ADR_PAL_INDEX
	and #$10
	bne +
	
	; black background
	lda #$00
	sta VIDEO_ADR_PAL_COLOR
	sta VIDEO_ADR_PAL_COLOR
	sta VIDEO_ADR_PAL_COLOR
	bra ++
	
+	; white backgound
	lda #$3F
	sta VIDEO_ADR_PAL_COLOR
	sta VIDEO_ADR_PAL_COLOR
	sta VIDEO_ADR_PAL_COLOR
	
++	inx
	
	; foreground color
	txa
	sta VIDEO_ADR_PAL_INDEX
	
	; foreground.r
	txa
	and #$80
	bne +
	
	lda #$3F
	sta VIDEO_ADR_PAL_COLOR
	bra ++
	
+	lda #$00
	sta VIDEO_ADR_PAL_COLOR
	
++	; foreground.g
	txa
	and #$40
	bne +
	
	lda #$3F
	sta VIDEO_ADR_PAL_COLOR
	bra ++
	
+	lda #$00
	sta VIDEO_ADR_PAL_COLOR
	
++	; foreground.b
	txa
	and #$20
	bne +
	
	lda #$3F
	sta VIDEO_ADR_PAL_COLOR
	bra ++
	
+	lda #$00
	sta VIDEO_ADR_PAL_COLOR	
	
++	inx
	
	bne -
					
	rts	
	
	
; --- Console clear screen
; @noparams
; Clears contents of VRAM to $20 (space).
console_clear_screen:
	phb
	bank #VIDEO_RAM_BANK 
	
	; init the row pointer to the start of row 29
	clr VIDEO_RAM_A0_L
	lda #$74
	sta VIDEO_RAM_A0_H
	
	; clear one row	
--	lda #$20 ; space character
	ldx VIDEO_RAM_LEFT_MARGIN
-	sta (VIDEO_RAM_A0),X
	inx
	inx
	bpl -	
	
	; subtract 4 from high byte of pointer to go back 8 rows = 1 text line
	lda VIDEO_RAM_A0_H
	sec
	sbc #4 
	sta VIDEO_RAM_A0_H
	bcs --
	
	plb
	rts
	
	
; --- Console home cursor
; @noparams
console_home_cursor:
	phb
	bank #VIDEO_RAM_BANK
	lda VIDEO_RAM_LEFT_MARGIN
	sta VIDEO_RAM_CURSOR_PTR_L	
	clr VIDEO_RAM_CURSOR_PTR_H
	plb
	rts
	
	
; --- Console cursor off
; @noparams
console_cursor_off:
	phb
	bank #VIDEO_RAM_BANK
	ldx #$00
	lda (VIDEO_RAM_CURSOR_PTR),X
	and #$7F
	sta (VIDEO_RAM_CURSOR_PTR),X
	plb
	rts
	
	
; --- Console cursor on
; @noparams
console_cursor_on:
	phb
	bank #VIDEO_RAM_BANK
	ldx #$00
	lda (VIDEO_RAM_CURSOR_PTR),X
	ora #$80
	sta (VIDEO_RAM_CURSOR_PTR),X
	plb
	rts
	
			
	
; --- Console write string 
; Print the null-terminated string pointed to by CONSOLE_ADR_A0 to the console 
console_write_string:
	ldx #$00
	
-	lda (CONSOLE_ADR_A0),X
	beq +
	cmp #CONSOLE_RETURN
	beq +
	cmp #CONSOLE_LINE_FEED
	beq +
	
	phx
	jsr console_write_char
	plx
	inx
	
	; terminate after 256 chars, even if NULL is not reached
	bne -

+	rts

			
; --- Console write string line
; Print the null-terminated string pointed to by CONSOLE_ADR_A0 to the console,
; followed by a newline. 
console_write_string_line:
	ldx #$00
	
-	lda (CONSOLE_ADR_A0),X
	beq +
	cmp #CONSOLE_RETURN
	beq +
	cmp #CONSOLE_LINE_FEED
	beq +
	
	phx
	jsr console_write_char
	plx
	inx
	
	; terminate after 256 chars, even if NULL is not reached
	bne -

+	lda #CONSOLE_RETURN
	jsr console_write_char
	rts


; --- Console write character
; @param0 - A one byte ASCII character value, to be displayed at the current cursor position.
; The display will be scrolled if necessary. Return is handled as well.	
console_write_char:	
	pha
	jsr console_cursor_off
	pla
	
	; return character?
	cmp #CONSOLE_RETURN
	bne +
	jsr console_advance_row
	jsr console_cursor_on
	rts
	
	; backspace?
+	cmp #CONSOLE_BACKSPACE
	bne +
	jsr console_backspace
	jsr console_cursor_on
	rts
		
+	; store char at cursor position
	phb
	bank #VIDEO_RAM_BANK
	ldx #$00
	sta (VIDEO_RAM_CURSOR_PTR),X
	plb
	
	; advance cursor and return	
	jsr console_advance_cursor
	jsr console_cursor_on
	rts

	
; --- Console advance cursor
; @noparams
console_advance_cursor:	
	phb
	bank #VIDEO_RAM_BANK
	
	lda VIDEO_RAM_CURSOR_PTR_L
	inc
	inc
	sta VIDEO_RAM_CURSOR_PTR_L
	and #$7F
	plb
	bne +
	
	jsr console_advance_row
	
+	rts
	
	
; --- Console advance row
; @noparams
console_advance_row:	
	phb
	bank #VIDEO_RAM_BANK
	
	; row low byte
	lda VIDEO_RAM_LEFT_MARGIN
	sta VIDEO_RAM_CURSOR_PTR_L
	
	; row high byte
	lda VIDEO_RAM_CURSOR_PTR_H
	clc
	adc #4
	sta VIDEO_RAM_CURSOR_PTR_H	
	
	; check for end of screen
	cmp #$78
	bne +
	jsr console_scroll	
	
+	plb
	rts
	

; --- Console backup cursor
; @noparams
console_backup_cursor:	
	phb
	bank #VIDEO_RAM_BANK
	
	; if already at row 0, column 0, then don't backup
	lda VIDEO_RAM_LEFT_MARGIN
	cmp VIDEO_RAM_CURSOR_PTR_L
	bne +
	cmp VIDEO_RAM_CURSOR_PTR_H
	bne +
	plb
	rts
	
+	lda VIDEO_RAM_CURSOR_PTR_L
	cmp VIDEO_RAM_LEFT_MARGIN
	bne +
	
	; wrap to previous line
	lda #$7E
	sta VIDEO_RAM_CURSOR_PTR_L
	
	; previous row
	; row high byte
	lda VIDEO_RAM_CURSOR_PTR_H
	sec
	sbc #4
	sta VIDEO_RAM_CURSOR_PTR_H	
	bra ++
	
+	; backup within the same line
	dec
	dec
	sta VIDEO_RAM_CURSOR_PTR_L
	
++	plb
	rts
	
		
; --- Console backspace
; @noparams
; Moves the cursor backwards by one logical position. Prints a space at the new position.
console_backspace:
	jsr console_backup_cursor

	phb
	bank #VIDEO_RAM_BANK
	
	lda #$20
	ldx #$00
	sta (VIDEO_RAM_CURSOR_PTR),X
	
	plb
	rts
		
	
; --- Console scroll
; @noparams
; Scrolls the contents of the console up by one line, and sets the cursor to the beginning
; of the last line.
console_scroll:
	phb
	bank #VIDEO_RAM_BANK 
	
	; init the dest row pointer to the start of the first row
	clr VIDEO_RAM_A0_L
	lda VIDEO_RAM_TOP_MARGIN
	sta VIDEO_RAM_A0_H

	; init the source row pointer to the start of the second row
	clr VIDEO_RAM_A1_L
	lda VIDEO_RAM_TOP_MARGIN
	clc
	adc #$04
	sta VIDEO_RAM_A1_H
		
	; copy one row	
--	ldx #$7E
-	lda (VIDEO_RAM_A1),X
	sta (VIDEO_RAM_A0),X
	dex
	dex
	bpl -	
	
	; add 4 to the high byte of both pointers to go forward 8 rows = 1 text line
	lda VIDEO_RAM_A0_H
	clc
	adc #4 
	sta VIDEO_RAM_A0_H
	
	lda VIDEO_RAM_A1_H
	clc
	adc #4 
	sta VIDEO_RAM_A1_H
	
	; stop at source row 30
	cmp #$78
	bne --
	
	; fill row 29 with spaces
	lda #$20 ; space character
	ldx VIDEO_RAM_LEFT_MARGIN
-	sta (VIDEO_RAM_A0),X
	inx
	inx
	bpl -	
	
	; set the cursor to the start of row 29
	lda VIDEO_RAM_LEFT_MARGIN
	sta VIDEO_RAM_CURSOR_PTR_L
	lda #$74
	sta VIDEO_RAM_CURSOR_PTR_H
	
	plb
	rts
	
	
	
; --- Console write hex nibble
; @param0 - A value less than 16, to be displayed as a signle hex digit at the current cursor position.
; The display will be scrolled if necessary.	
console_write_hex_nibble:
	cmp #10
	bmi +
	
	; A-F
	clc
	adc #55
	jsr console_write_char
	bra ++
		
	; 0-9
+	clc
	adc #'0'
	jsr console_write_char
	
++  rts

	
; --- Console write hex byte
; @param0 - A one byte value, to be displayed as two hex digits, beginning at the current cursor position.
; The display will be scrolled if necessary.	
console_write_hex:	
	; store A
	pha
	
	; A = A >> 4
	; ROL * 5 followed by AND should be faster than LSR * 4
	rol
	rol
	rol
	rol
	rol
	and #$F
	jsr console_write_hex_nibble
	
	; restore A, A = A & 0xF
	pla
	and #$F
	jsr console_write_hex_nibble
	
	rts
	